import os
from os import sep
from glob import glob
import pandas as pd
import numpy as np
# from tqdm import tqdm
from datetime import datetime, timedelta
from microt_prompt_matrix import utils

target_file_pattern = 'phone_watch_daily_report_clean*.csv'

def validate_dates_before_after(intermediate_participant_path, date_in_study):
    validated_date_list = []

    # target date
    date_folder_path = intermediate_participant_path + sep + date_in_study
    target_date_log_paths = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
    if len(target_date_log_paths) == 0:
        print("No daily report file on {}".format(date_in_study))
    else:

        # 1 day before target date
        date_format = "%Y-%m-%d"
        one_date_before_datetime = datetime.strptime(date_in_study, date_format).date() - timedelta(days=1)
        one_date_before = one_date_before_datetime.strftime(date_format)
        date_folder_path = intermediate_participant_path + sep + one_date_before
        one_day_before_log_paths = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
        if len(one_day_before_log_paths) != 0:
            validated_date_list.append(one_date_before)

        # target date
        validated_date_list.append(date_in_study)

        # 1 day after target date
        date_format = "%Y-%m-%d"
        one_date_after_datetime = datetime.strptime(date_in_study, date_format).date() + timedelta(days=1)
        one_date_after = one_date_after_datetime.strftime(date_format)
        date_folder_path = intermediate_participant_path + sep + one_date_after
        one_day_after_log_paths = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
        if len(one_day_after_log_paths) != 0:
            validated_date_list.append(one_date_after)


    return validated_date_list

def combine_intermediate_file(intermediate_participant_path, date_in_study):
    df_logs_combined = pd.DataFrame()
    participant_id = utils.extract_participant_id.extract_participant_id(intermediate_participant_path)

    # generate date range where date folder exists (sharable code in utils)
    validated_date_list = validate_dates_before_after(intermediate_participant_path, date_in_study)
    if len(validated_date_list) == 0:
        print("Cannot find logs file around {}".format(date_in_study))
        return df_logs_combined

    for date in validated_date_list:
        date_folder_path = intermediate_participant_path + sep + date
        csv_path_list = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name

        csv_path = csv_path_list[0]
        df_day = pd.read_csv(csv_path)
        if df_day.shape[0] > 0:
            df_day['Participant_ID'] = [participant_id] * df_day.shape[0]
            df_logs_combined = pd.concat([df_logs_combined, df_day])


    converter = lambda x: datetime.strptime(x, "%Y-%m-%d")
    df_logs_combined = df_logs_combined.dropna(subset=['date'])
    df_logs_combined.reset_index(inplace=True, drop=True)

    df_logs_combined['Date_Datetime'] = pd.Series(map(converter, df_logs_combined["date"]))
    df_logs_combined['Date'] = df_logs_combined['Date_Datetime'].dt.date

    return df_logs_combined


def match_feature(prompt_local_datetime_series, df_logs_combined):

    prompt_hour_list = [int(x.hour) for x in prompt_local_datetime_series]
    prompt_date_datetime_list = [x.date() for x in prompt_local_datetime_series]
    prompt_feature_df = pd.DataFrame({"Prompt_Hour": prompt_hour_list, "Prompt_Date_Datetime": prompt_date_datetime_list})

    print("     --- start matching\n")
    matched_waking_hours_list = []
    matched_sleep_hours_list = []
    for idx in prompt_feature_df.index:

        prompt_hour = prompt_feature_df['Prompt_Hour'][idx]
        # print(prompt_hour)

        if prompt_hour > 4:
            prompt_date = prompt_feature_df['Prompt_Date_Datetime'][idx]
        else:
            prompt_date = prompt_feature_df['Prompt_Date_Datetime'][idx] - timedelta(days=1)

        df_logs_combined_filtered = df_logs_combined[df_logs_combined['Date'] == prompt_date]

        if df_logs_combined_filtered.shape[0] == 0:
            waking_hour = np.nan
            sleep_hour = np.nan
            print("No wake and sleep hour found for {}".format(prompt_date))
        else:
            df_logs_combined_filtered.reset_index(drop=True, inplace=True)
            waking_hour = list(df_logs_combined_filtered["current_wake_time"])[0]
            sleep_hour = list(df_logs_combined_filtered["current_sleep_time"])[0]

        matched_waking_hours_list.append(waking_hour)
        matched_sleep_hours_list.append(sleep_hour)

    return matched_waking_hours_list, matched_sleep_hours_list


def transform(waking_hours_column):
    # return [1 if x >= 15 else 0 for x in waking_hours_column]
    return waking_hours_column


def create_column(prompt_local_datetime_series, intermediate_participant_path, date_in_study):
    print("\n> start generating the feature: Waking hours ")

    # Read, parse and combine related intermediate file
    df_logs_combined = combine_intermediate_file(intermediate_participant_path, date_in_study)

    if df_logs_combined.shape[0] > 0:
        # Match the combined parsed intermediate file with prompt feature data frame
        waking_hours_column, sleep_hours_column = match_feature(prompt_local_datetime_series, df_logs_combined)
    else:
        waking_hours_column = np.nan * len(prompt_local_datetime_series)
        sleep_hours_column = np.nan * len(prompt_local_datetime_series)

    print("     --- success")


    return waking_hours_column, sleep_hours_column


import os
from os import sep
from glob import glob
import pandas as pd
import numpy as np
import bisect
from datetime import datetime, timedelta
from microt_prompt_matrix import utils
from microt_prompt_matrix.utils.get_time_diff import *

target_file_pattern = 'watch_accelerometer_mims*.csv'


def combine_intermediate_file(intermediate_participant_path, date_in_study):
    df_logs_combined = pd.DataFrame()
    participant_id = utils.extract_participant_id.extract_participant_id(intermediate_participant_path)

    # generate date range where date folder exists (sharable code in utils)
    # validated_date_list = validate_dates_before_after(intermediate_participant_path, date_in_study)
    # if len(validated_date_list) == 0:
    #     print("Cannot find logs file around {}".format(date_in_study))
    #     return df_logs_combined

    for date in [date_in_study]:
        date_folder_path = intermediate_participant_path + sep + date
        csv_path_list = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
        if len(csv_path_list) == 0:
            print("Cannot find logs file around {}".format(date_in_study))
            return df_logs_combined

        csv_path = csv_path_list[0]

        try:
            df_day = pd.read_csv(csv_path)
            if df_day.shape[0] > 0:
                df_day['Participant_ID'] = [participant_id] * df_day.shape[0]
                df_logs_combined = pd.concat([df_logs_combined, df_day])
        except pd.errors.EmptyDataError:
            print("pandas.errors.EmptyDataError (phone mims) : " + csv_path)
            return df_logs_combined

    converter = lambda x: datetime.strptime(x, "%Y-%m-%d %H:%M:%S.%f") if ("." in x) else datetime.strptime(x,
                                                                                                            "%Y-%m-%d %H:%M:%S")

    df_logs_combined = df_logs_combined.dropna(subset=['HEADER_TIME_STAMP'])
    df_logs_combined.reset_index(inplace=True, drop=True)
    df_logs_combined["HEADER_TIME_STAMP"] = [x.split(" ")[0] + " " + x.split(" ")[1] for x in
                                             list(df_logs_combined["HEADER_TIME_STAMP"])]

    df_logs_combined['TIME_DATETIME'] = pd.Series(map(converter, df_logs_combined["HEADER_TIME_STAMP"]))
    # df_logs_combined['Date'] = df_logs_combined['TIME_DATETIME'].dt.date

    return df_logs_combined


def find_closest_time(prompt_time, subset_time_list):
    pos = bisect.bisect_left(subset_time_list, prompt_time)
    # closet_time = min(subset_time_list[max(0, i - 1): i + 2], key=lambda t: -(prompt_time - t))

    return pos


def get_mims_summary(sec_before, prompt_time, closest_times, df_logs_combined):
    mims_summary = 0
    num_readings = 0
    start_time = None
    for matched_idx in closest_times:
        closest_time = df_logs_combined.loc[matched_idx, "TIME_DATETIME"]
        if get_min_diff(prompt_time, closest_time) <= (sec_before // 60 + 1):
            mims_min = df_logs_combined.loc[matched_idx, "MIMS_UNIT"]
            if mims_min != -0.01:  # mims value -0.01 means cannot be computed
                mims_summary += mims_min
                num_readings += 1
            start_time = closest_time
        else:
            if start_time is None:
                mims_summary = "OB"
            break
    return mims_summary, num_readings, start_time


def match_feature(prompt_local_datetime_series, df_logs_combined):
    print("     --- start matching")
    # Aggregating minutes before prompt time
    sec_before_list = [60 * x for x in list(range(1, 11))]

    mims_summary_list = []
    start_time_list = []
    readings_list = []

    for idx in prompt_local_datetime_series.index:
        matched_mims_summary_list = []
        matched_start_time_list = []
        matched_readings_list = []

        prompt_time = prompt_local_datetime_series[idx]
        # prompt_date = prompt_time.date()

        if df_logs_combined.shape[0] == 0:
            matched_mims_summary_list = [np.nan] * len(sec_before_list)
            matched_readings_list = [np.nan] * len(sec_before_list)
            matched_start_time_list = [np.nan] * len(sec_before_list)
        else:
            subset_time_list = list(df_logs_combined["TIME_DATETIME"])
            pos = find_closest_time(prompt_time, subset_time_list)

            for sec_before in sec_before_list:
                closest_times = []
                for i in range(sec_before):
                    closest_times.append(pos - 1 - i)
                closest_times = [x for x in closest_times if x >= 0]
                mims_summary, readings, start_time = get_mims_summary(sec_before, prompt_time, closest_times,
                                                                      df_logs_combined)
                matched_mims_summary_list.append(mims_summary)
                matched_readings_list.append(readings)
                matched_start_time_list.append(start_time)
        mims_summary_list.append(matched_mims_summary_list)
        readings_list.append(matched_readings_list)
        start_time_list.append(matched_start_time_list)

    mims_summary_df = pd.DataFrame(list(map(np.ravel, mims_summary_list)))
    readings_df = pd.DataFrame(list(map(np.ravel, readings_list)))
    start_time_df = pd.DataFrame(list(map(np.ravel, start_time_list)))

    df = pd.DataFrame()
    for sec_before in sec_before_list:
        col = int(sec_before // 60 - 1)
        df = pd.concat([df, mims_summary_df[col].rename("mims_summary_" + str(sec_before // 60) + "min")], axis=1)
        df = pd.concat([df, readings_df[col].rename("num_readings_" + str(sec_before // 60) + "min")], axis=1)
        df = pd.concat([df, start_time_df[col].rename("start_time_" + str(sec_before // 60) + "min")], axis=1)

    return df


def create_column(prompt_local_datetime_series, intermediate_participant_path, date_in_study):
    print("\n> start generating the feature: MIMS ")

    # Read, parse and combine related intermediate file
    df_logs_combined = combine_intermediate_file(intermediate_participant_path, date_in_study)

    # Match the combined parsed intermediate file with prompt feature data frame
    df = match_feature(prompt_local_datetime_series, df_logs_combined)

    return df


if __name__ == "__main__":
    pass



import pandas as pd
import numpy as np
from datetime import datetime


def convert_time_str_to_datetime(time_str_series):
    converter = lambda x: datetime.strptime(x, "%Y-%m-%d %H:%M:%S") if (type(x) == str) else x
    time_datetime_list = [x.split(" ")[0] + " " + x.split(" ")[1].split(".")[0] if type(x) == str else x for x in
                          list(time_str_series)]

    time_datetime_list = pd.Series(map(converter, time_datetime_list))

    return time_datetime_list


def create_column(feature_df, parts_num):
    print("\n> start generating the feature: Parts of waking hours")
    wake_time_datetime_list = convert_time_str_to_datetime(feature_df["WAKE_TIME"])
    sleep_time_datetime_list = convert_time_str_to_datetime(feature_df["SLEEP_TIME"])
    prompt_time_datetime_list = feature_df["Actual_Prompt_Local_DateTime"]

    proximity_to_wake_time_column = [(y - x).total_seconds() / 60 if ((type(x) != float) and (type(y) != float)) else np.nan
                                     for x, y in zip(wake_time_datetime_list, prompt_time_datetime_list)]
    proximity_to_sleep_time_column = [(y - x).total_seconds() / 60 if ((type(x) != float) and (type(y) != float)) else np.nan
                                      for x, y in zip(prompt_time_datetime_list, sleep_time_datetime_list)]
    wake_duration = [(y - x).total_seconds() / 60 if ((type(x) != float) and (type(y) != float)) else np.nan for x, y in
                     zip(wake_time_datetime_list, sleep_time_datetime_list)]
    duration_per_part = [x // parts_num if x == x else x for x in wake_duration]
    parts_of_waking_hour_column = [y // x if ((x == x) and (y == y)) else np.nan for x, y in
                                   zip(duration_per_part, proximity_to_wake_time_column)]

    print("     --- success")

    return parts_of_waking_hour_column, proximity_to_wake_time_column, proximity_to_sleep_time_column



import os
import json
import time
from glob import glob
import pandas as pd
import numpy as np
from math import degrees

def add_start_date(participants_folder_path_list, info_dict):
    for p_folder_path in participants_folder_path_list:
        participant_id = p_folder_path.split(os.sep)[-1]
        print(participant_id)
        info_dict[participant_id] = dict()
        date_list = list(os.listdir(p_folder_path))
        date_list_r = date_list[::-1]
        for date_folder in date_list_r:
            target_list = list(glob(os.path.join(p_folder_path, date_folder, "phone_watch_daily_report*")))
            if len(target_list) > 0:
                df_daily = pd.read_csv(target_list[0])
                start_date_list = df_daily['start_date'].unique()
                for start_date in start_date_list:
                    if type(start_date) == str: # str or nan
                        info_dict[participant_id]['start_date'] = start_date
                        print(start_date)
                        break
                else:
                    continue # only executed if the inner loop did NOT break
                break # only executed if the inner loop DID break
    return info_dict

def read_json(json_path):
    with open(json_path) as f:
        data = json.load(f)
    return data

def convert_radians_to_degrees(loc):
    if abs(loc) < 7:
        loc = degrees(loc)
    return round(loc, 5)

def add_home_locations(lsf_json_folder_path, info_dict):
    for participant_id in info_dict:
        lsf_json_path = os.path.join(lsf_json_folder_path, participant_id, "lsf.json")
        if not os.path.exists(lsf_json_path):
            continue
        else:
            lsf_json = read_json(lsf_json_path)
            labels = lsf_json["label"]
            bboxes = lsf_json["bbox"]
            home_coord_list = []
            for clusterID in labels:
                if "Home" in labels[clusterID]:
                    if clusterID in bboxes:
                        bbox = bboxes[clusterID]
                        home_lat = convert_radians_to_degrees(np.mean([bbox[0], bbox[1]]))
                        home_lon = convert_radians_to_degrees(np.mean([bbox[2], bbox[3]]))
                        home_coord = [home_lat, home_lon]
                        home_coord_list.append(home_coord)
            info_dict[participant_id]['home_loc'] = home_coord_list

    return info_dict


def create_json(intermediate_participant_path, output_dir_path, lsf_json_folder_path):
    print("Creating new participant info json.")
    info_dict = dict()
    misc_folder_name = "misc"
    information_dict_name = "info.json"
    folder_path = os.path.join(output_dir_path, misc_folder_name)
    information_dict_path = os.path.join(folder_path, information_dict_name)
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)

    # start_date
    intermediate_folder_path = os.path.abspath(os.path.join(intermediate_participant_path, os.pardir))
    participants_folder_path_list = list(glob(os.path.join(intermediate_folder_path, "*@timestudy_com")))
    info_dict = add_start_date(participants_folder_path_list, info_dict)

    # home location
    info_dict = add_home_locations(lsf_json_folder_path, info_dict)

    # dump json file
    with open(information_dict_path, 'w') as f:
        json.dump(info_dict, f)

    print("Creating new participant info json. (Done)")
    return info_dict


def execute(intermediate_participant_path, output_dir_path, lsf_json_folder_path):

    misc_folder_name = "misc"
    information_dict_name = "info.json"
    folder_path = os.path.join(output_dir_path, misc_folder_name)
    information_dict_path = os.path.join(folder_path, information_dict_name)

    if not os.path.exists(information_dict_path): # when the json needs to be updated
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        info_dict = create_json(intermediate_participant_path, output_dir_path, lsf_json_folder_path)
    else:
        # Load existing json file
        # if os.path.exists(information_dict_path):
        #     with open(information_dict_path) as f:
        #         info_dict = json.load(f)
        # else: # Create if info file doesn't exist
        #     info_dict = create_json(intermediate_participant_path, output_dir_path)
        with open(information_dict_path) as f:
            info_dict = json.load(f)

    return info_dict



if __name__ == "__main__":
    output_dir_path = r"C:\Users\jixin\Desktop\test_result\compliance"
    intermediate_participant_path = r"G:\preprocessed_cluster\intermediate_file\sharpnessnextpouch@timestudy_com"
    start = time.time()
    info_dict = execute(intermediate_participant_path, output_dir_path)
    end = time.time()
    print(info_dict)
    print("Runtime : {}".format(round(end-start, 2)))


"""
Functionality: Parse the content of .../logs-watch/PromptResponses.log.csv for individual participant into formatted
matrices day by day and save as csv file in the directory .../participant_id@timestudy_com/appended_response.

Usage: This script is intended mainly to be called by uema_all_features_dataframe.py.

Input: python preprocess_promptResponse_raw.py [PARTICIPANT_ID] [microT_root_path] [start_date]
e.g., python preprocess_promptResponse_raw.py aditya4_internal .../MICROT 2020-01-01

Output: csv file for each day of the participant, with each row indicating a prompted question and response
e.g., aditya4_internal_uEMA_****-**-**.csv in .../aditya4_internal@timestudy_com/appended_response
"""
import os
import sys
import numpy as np
import pandas.errors

from microt_prompt_matrix.utils.convert_timestamp import *
from glob import glob

columns_list = ["Participant_ID", "Initial_Prompt_Date", "Prompt_Type", "Study_Mode", "Initial_Prompt_Local_Time",
                "Initial_Prompt_UnixTime",
                "Initial_Prompt_UTC_Offset ",
                "Answer_Status", 'Question_Set_Completion_Local_Time', 'Question_Set_Completion_UnixTime',
                "Question_X_Answer_Unixtime"]

prompt_type_excluded = ['None_Sleep', 'None_Empty']


def extract_reprompt(df_pr_watch):
    prompt_arrays = []
    for idx in df_pr_watch.index:
        row = df_pr_watch.loc[idx]

        new_reprompt_row = []
        new_reprompt_row.append(row["Participant_ID"])
        new_reprompt_row.append(row["Initial_Prompt_Date"])
        new_reprompt_row.append(row["Prompt_Type"])
        new_reprompt_row.append(row["Study_Mode"])
        new_reprompt_row.append(row["Initial_Prompt_Local_Time"])
        new_reprompt_row.append(row["Answer_Status"])
        new_reprompt_row.append(row["Initial_Prompt_Local_Time"])
        new_reprompt_row.append(row["Question_X_Answer_Unixtime"])
        new_reprompt_row.append(row["Initial_Prompt_UTC_Offset "])
        new_reprompt_row.append(0)

        prompt_arrays.append(new_reprompt_row)

    df_prompt = pd.DataFrame(prompt_arrays)
    if df_prompt.shape[0] > 0:
        df_prompt.columns = ["Participant_ID", "Initial_Prompt_Date", "Prompt_Type", "Study_Mode",
                             "Initial_Prompt_Local_Time", "Answer_Status",
                             "Actual_Prompt_Local_Time", "First_Question_Completion_Unixtime", "UTC_Offset",
                             "Reprompt_Num"]

    return df_prompt


def clean_response_files(intermediate_participant_path, date_in_study, participant_id):
    # Check if date folder exists
    date_folder_path = os.path.join(intermediate_participant_path, date_in_study)
    # date_exist = validate_date(date_folder_path)
    # if not date_exist:
    #     print("Cannot find date folder for {}".format(date_in_study))
    #     raise Exception("Cannot find date folder for {}".format(date_in_study))

    # Check if phone_promptresponse file exists for this date
    csv_path = list(glob(os.path.join(date_folder_path, 'watch_promptresponse_clean*.csv')))
    if len(csv_path) == 0:
        print(date_folder_path)
        print("No intermediate watch prompt response csv file on {}".format(date_in_study))
        quit()

    # read and filter prompt type
    try:
        df_raw = pd.read_csv(csv_path[0])
    except pandas.errors.EmptyDataError:
        raise Exception("pandas.errors.EmptyDataError (watch prompt response) : " + csv_path[0])

    if df_raw.shape[0] == 0:
        print("Empty watch prompt response file : " + csv_path[0])
        quit()

    if "Question_X_Answer_Unixtime" not in df_raw:
        print("All watch prompts are never prompted {}".format(date_folder_path))
        quit()

    df_raw = df_raw[columns_list]

    df_raw = df_raw[~df_raw.Prompt_Type.isin(prompt_type_excluded)]
    df_raw = df_raw[df_raw['Initial_Prompt_UnixTime'] != -1]
    df_raw.reset_index(drop=True, inplace=True)

    # extract reprompts and treated them as separate prompts
    df_reprompt = extract_reprompt(df_raw)

    # add line number as prompt key. First prompt = 1.
    df_reprompt["Line"] = np.arange(1, len(df_reprompt) + 1)

    if df_reprompt.shape[0] == 0:
        print("Empty watch prompt response file after filtering : " + csv_path[0])
        quit()

    # add participant id column
    # df_reprompt.insert(loc=0, column='Participant_ID', value=participant_id)

    # time zone processing
    df_reprompt.reset_index(drop=True, inplace=True)
    Prompt_Local_Time_list = []
    for idx in df_reprompt.index:
        prompt_local_time_timezone = df_reprompt['Actual_Prompt_Local_Time'][idx]
        if type(prompt_local_time_timezone) == float:
            Prompt_Local_Time_list.append(prompt_local_time_timezone)
            print("float : {}".format(prompt_local_time_timezone))
            continue
        prompt_local_time_timezone_split = prompt_local_time_timezone.split(" ")
        prompt_local_time_timezone_split_len = len(prompt_local_time_timezone_split)
        if prompt_local_time_timezone_split_len == 3:
            Prompt_Local_Time_list.append(
                prompt_local_time_timezone_split[0] + " " + prompt_local_time_timezone_split[1])
        elif prompt_local_time_timezone_split_len == 6:
            b = [1, 2, 3, 5]
            subset_time = [prompt_local_time_timezone_split[i] for i in b]
            Prompt_Local_Time_list.append(" ".join(subset_time))
        else:
            print("Prompt response time zone error : " + intermediate_participant_path + "//" + date_in_study)
            raise Exception("Unable to parse Actual_Prompt_Local_Time")

    prompt_local_time_timezone = df_reprompt['Actual_Prompt_Local_Time'][0]
    prompt_local_time_timezone_split = prompt_local_time_timezone.split(" ")
    prompt_local_time_timezone_split_len = len(prompt_local_time_timezone_split)
    if prompt_local_time_timezone_split_len == 3:
        converter = lambda x: datetime.strptime(x, "%Y-%m-%d %H:%M:%S") if (type(x) != float) else np.nan
        prompt_dateobject_column = list(map(converter, Prompt_Local_Time_list))
    elif prompt_local_time_timezone_split_len == 6:
        converter = lambda x: datetime.strptime(x, "%b %d %H:%M:%S %Y") if (type(x) != float) else np.nan
        prompt_dateobject_column = list(map(converter, Prompt_Local_Time_list))
    else:
        raise Exception("Unable to parse Actual_Prompt_Local_Time")

    # drop float type in Actual_Prompt_Local_Time
    df_reprompt['Actual_Prompt_Local_DateTime'] = prompt_dateobject_column
    df_reprompt = df_reprompt[df_reprompt['Actual_Prompt_Local_DateTime'].notnull()]
    df_reprompt.reset_index(inplace=True, drop=True)

    return df_reprompt


if __name__ == "__main__":
    p_id = sys.argv[1]
    intermediate_root_path = sys.argv[2]
    feature_save_path = sys.argv[3]
    start_date = sys.argv[4]
    end_date = sys.argv[5]
    clean_response_files(p_id, intermediate_root_path, feature_save_path, start_date, end_date)



import os
from os import sep
from glob import glob
import pandas as pd
import numpy as np
import bisect
from datetime import datetime, timedelta
from microt_prompt_matrix import utils
from microt_prompt_matrix.utils.get_time_diff import *

target_file_pattern_mims = 'watch_accelerometer_mims_clean*.csv'
target_file_pattern_mims_minute = 'watch_accelerometer_mims_minute*.csv'
target_file_pattern_swan_minute = 'watch_accelerometer_swan_minute*.csv'


def validate_dates_before_after(intermediate_participant_path, date_in_study, target_file_pattern):
    validated_date_list = []

    # target date
    date_folder_path = intermediate_participant_path + sep + date_in_study
    target_date_log_paths = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
    if len(target_date_log_paths) == 0:
        print("No battery daily file on {}".format(date_in_study))
    else:

        # 1 day before target date
        date_format = "%Y-%m-%d"
        one_date_before_datetime = datetime.strptime(date_in_study, date_format).date() - timedelta(days=1)
        one_date_before = one_date_before_datetime.strftime(date_format)
        date_folder_path = intermediate_participant_path + sep + one_date_before
        one_day_before_log_paths = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
        if len(one_day_before_log_paths) != 0:
            validated_date_list.append(one_date_before)

        # target date
        validated_date_list.append(date_in_study)

        # 1 day after target date
        date_format = "%Y-%m-%d"
        one_date_after_datetime = datetime.strptime(date_in_study, date_format).date() + timedelta(days=1)
        one_date_after = one_date_after_datetime.strftime(date_format)
        date_folder_path = intermediate_participant_path + sep + one_date_after
        one_day_after_log_paths = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
        if len(one_day_after_log_paths) != 0:
            validated_date_list.append(one_date_after)

    return validated_date_list


def combine_intermediate_mims_file(intermediate_participant_path, date_in_study, target_file_pattern):
    df_logs_combined = pd.DataFrame()
    participant_id = utils.extract_participant_id.extract_participant_id(intermediate_participant_path)

    # generate date range where date folder exists (sharable code in utils)
    validated_date_list = validate_dates_before_after(intermediate_participant_path, date_in_study, target_file_pattern)
    if len(validated_date_list) == 0:
        print("Cannot find logs file around {}".format(date_in_study))
        return df_logs_combined

    for date in validated_date_list:
        date_folder_path = intermediate_participant_path + sep + date
        csv_path_list = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
        if len(csv_path_list) == 0:
            print("Cannot find logs file around {}".format(date))
            # return df_logs_combined
            continue

        csv_path = csv_path_list[0]

        try:
            df_day = pd.read_csv(csv_path)
            if df_day.shape[0] > 0:
                df_day['Participant_ID'] = [participant_id] * df_day.shape[0]
                df_logs_combined = pd.concat([df_logs_combined, df_day])
        except pd.errors.EmptyDataError:
            print("pandas.errors.EmptyDataError (phone mims) : " + csv_path)
            continue

    if len(df_logs_combined) > 0:
        converter = lambda x: datetime.strptime(x, "%Y-%m-%d %H:%M:%S.%f") if ("." in x) else datetime.strptime(x,
                                                                                                                "%Y-%m-%d %H:%M:%S")

        df_logs_combined = df_logs_combined.dropna(subset=['HEADER_TIME_STAMP'])
        df_logs_combined.reset_index(inplace=True, drop=True)
        df_logs_combined["HEADER_TIME_STAMP"] = [x.split(" ")[0] + " " + x.split(" ")[1] for x in
                                                 list(df_logs_combined["HEADER_TIME_STAMP"])]

        df_logs_combined['TIME_DATETIME'] = pd.Series(map(converter, df_logs_combined["HEADER_TIME_STAMP"]))
    # df_logs_combined['Date'] = df_logs_combined['TIME_DATETIME'].dt.date

    return df_logs_combined


def combine_intermediate_mims_swan_minute_file(intermediate_participant_path, date_in_study, target_file_pattern):
    df_logs_combined = pd.DataFrame()
    participant_id = utils.extract_participant_id.extract_participant_id(intermediate_participant_path)

    # generate date range where date folder exists (sharable code in utils)
    validated_date_list = validate_dates_before_after(intermediate_participant_path, date_in_study, target_file_pattern)
    if len(validated_date_list) == 0:
        print("Cannot find logs file around {}".format(date_in_study))
        return df_logs_combined

    for date in validated_date_list:
        date_folder_path = intermediate_participant_path + sep + date
        csv_path_list = sorted(glob(os.path.join(date_folder_path, target_file_pattern)))  # file name
        if len(csv_path_list) == 0:
            print("Cannot find logs file around {}".format(date))
            # return df_logs_combined
            continue

        csv_path = csv_path_list[0]

        try:
            df_day = pd.read_csv(csv_path)
            if df_day.shape[0] > 0:
                df_day['Participant_ID'] = [participant_id] * df_day.shape[0]
                df_logs_combined = pd.concat([df_logs_combined, df_day])
        except pd.errors.EmptyDataError:
            print("pandas.errors.EmptyDataError (phone mims) : " + csv_path)
            continue

    return df_logs_combined


def find_closest_time(prompt_time, subset_time_list):
    pos = bisect.bisect_left(subset_time_list, prompt_time)
    # closet_time = min(subset_time_list[max(0, i - 1): i + 2], key=lambda t: -(prompt_time - t))

    return pos


def get_mims_summary_before_prompt(sec_before, prompt_time, closest_times, df_logs_combined):
    mims_summary = 0
    num_readings = 0
    start_time = None
    for matched_idx in closest_times:
        closest_time = df_logs_combined.loc[matched_idx, "TIME_DATETIME"]
        if get_min_diff(prompt_time, closest_time) <= (sec_before // 60 + 1):
            # the matched start time should be within the minute range. This may not be true when the mims logs have gaps and return "OB".
            # Ex. time of a mims log that is 10 minutes before the prompt should be within 10 minutes range of prompt time.
            mims_min = df_logs_combined.loc[matched_idx, "MIMS_UNIT"]
            if mims_min != -0.01:  # mims value -0.01 means cannot be computed
                mims_summary += mims_min
                num_readings += 1
            start_time = closest_time
        else:
            if start_time is None:
                mims_summary = "OB"
            break
    return mims_summary, num_readings, start_time


def get_mims_summary_after_prompt(sec_after, prompt_time, closest_times, df_logs_combined):
    mims_summary = 0
    num_readings = 0
    start_time = None
    for matched_idx in closest_times:
        closest_time = df_logs_combined.loc[matched_idx, "TIME_DATETIME"]
        if get_min_diff(prompt_time, closest_time) <= (sec_after // 60 + 1):
            # the matched start time should be within the minute range. This may not be true when the mims logs have gaps and return "OB".
            # Ex. time of a mims log that is 10 minutes before the prompt should be within 10 minutes range of prompt time.
            mims_min = df_logs_combined.loc[matched_idx, "MIMS_UNIT"]
            if mims_min != -0.01:  # mims value -0.01 means cannot be computed
                mims_summary += mims_min
                num_readings += 1
            start_time = closest_time
        else:
            if start_time is None:
                mims_summary = "OB"
            break
    return mims_summary, num_readings, start_time


def get_mims_summary_after_prompt_wearstate(wear_state, sec_after, prompt_time, df_mims_swan_minute_combined, date_in_study):
    mims_summary_wearstate = 0
    mims_wearstate_min = 0
    max_idx = len(df_mims_swan_minute_combined) - 1

    window_start_idices = df_mims_swan_minute_combined[
        (df_mims_swan_minute_combined.YEAR_MONTH_DAY == date_in_study) & (
                df_mims_swan_minute_combined.HOUR == prompt_time.hour) & (
                df_mims_swan_minute_combined.MINUTE == prompt_time.minute)].index
    if len(window_start_idices) > 0:
        window_start_idx = window_start_idices[0]
        min_num = sec_after // 60
        window_end_idx = window_start_idx + min_num
        if window_end_idx > max_idx:
            window_end_idx = max_idx
        for idx in range(window_start_idx, window_end_idx + 1):
            swan_pred = df_mims_swan_minute_combined.loc[idx, "SWAN_PREDICTION"]
            # print(swan_pred)
            if swan_pred != wear_state:
                continue
            mims_sample_num = df_mims_swan_minute_combined.loc[idx, "MIMS_SAMPLE_NUM"]
            if mims_sample_num > 0:
                mims_value = df_mims_swan_minute_combined.loc[idx, "MIMS_SUM"]
                mims_summary_wearstate += mims_value
                mims_wearstate_min += 1
    else:
        mims_summary_wearstate = "NF"
        mims_wearstate_min = "NF"

    return mims_summary_wearstate, mims_wearstate_min


def match_feature(prompt_local_datetime_series, df_mims_combined, df_mims_swan_minute_combined, date_in_study):
    print("     --- start matching")
    # Aggregating minutes before prompt time
    sec_before_list = [60 * x for x in list(range(1, 11))]  # 1-10 minute before prompt time
    sec_after_list = [30 * 60 * x for x in list(range(1, 7))]  # 30, 60, 90, 120, 150, 180 minute after prompt time

    mims_summary_before_list = []
    window_start_time_before_list = []
    window_readings_before_list = []

    mims_summary_after_list = []
    window_start_time_after_list = []
    window_readings_after_list = []

    mims_summary_wearstate_after_list = []
    wearstate_min_after_list = []

    for idx in prompt_local_datetime_series.index:
        # print(idx)
        matched_mims_summary_before_list = []
        matched_window_start_time_before_list = []
        matched_window_readings_before_list = []

        matched_mims_summary_after_list = []
        matched_window_start_time_after_list = []
        matched_window_readings_after_list = []

        matched_mims_summary_wearstate_after_list = []
        matched_wearstate_min_after_list = []

        prompt_time = prompt_local_datetime_series[idx]
        # prompt_date = prompt_time.date()

        if df_mims_combined.shape[0] == 0:
            matched_mims_summary_before_list = [np.nan] * len(sec_before_list)
            matched_window_readings_before_list = [np.nan] * len(sec_before_list)
            matched_window_start_time_before_list = [np.nan] * len(sec_before_list)

            matched_mims_summary_after_list = [np.nan] * len(sec_after_list)
            matched_window_readings_after_list = [np.nan] * len(sec_after_list)
            matched_window_start_time_after_list = [np.nan] * len(sec_after_list)

            matched_mims_summary_wearstate_after_list = [np.nan] * len(sec_after_list)
            matched_wearstate_min_after_list = [np.nan] * len(sec_after_list)

        else:
            subset_time_list = list(df_mims_combined["TIME_DATETIME"])
            pos = find_closest_time(prompt_time, subset_time_list)
            max_pos = len(subset_time_list) - 1

            # window before prompt
            for sec_before in sec_before_list:
                closest_times = []
                for i in range(sec_before):
                    closest_times.append(pos - 1 - i)
                closest_times = [x for x in closest_times if x >= 0]
                mims_summary, readings, start_time = get_mims_summary_before_prompt(sec_before, prompt_time,
                                                                                    closest_times,
                                                                                    df_mims_combined)
                matched_mims_summary_before_list.append(mims_summary)
                matched_window_readings_before_list.append(readings)
                matched_window_start_time_before_list.append(start_time)

            # window after prompt
            for sec_after in sec_after_list:
                # print(sec_after)
                closest_times = []
                for i in range(sec_after):
                    closest_times.append(pos + 1 + i)
                closest_times = [x for x in closest_times if x <= max_pos]

                # overall mims summary and sample num
                mims_summary, readings, start_time = get_mims_summary_after_prompt(sec_after, prompt_time,
                                                                                   closest_times,
                                                                                   df_mims_combined)

                # mims summary and sample num of sleep, wear, nonwear
                for wear_state in ["Wear", "Sleep", "Nonwear"]:
                    mims_summary_wearstate, min_wearstate = get_mims_summary_after_prompt_wearstate(wear_state,
                                                                                                    sec_after,
                                                                                                    prompt_time,
                                                                                                    df_mims_swan_minute_combined, date_in_study)

                    matched_mims_summary_wearstate_after_list.append(mims_summary_wearstate)
                    matched_wearstate_min_after_list.append(min_wearstate)

                matched_mims_summary_after_list.append(mims_summary)
                matched_window_readings_after_list.append(readings)
                matched_window_start_time_after_list.append(start_time)

        mims_summary_before_list.append(matched_mims_summary_before_list)
        window_readings_before_list.append(matched_window_readings_before_list)
        window_start_time_before_list.append(matched_window_start_time_before_list)

        mims_summary_after_list.append(matched_mims_summary_after_list)
        window_readings_after_list.append(matched_window_readings_after_list)
        window_start_time_after_list.append(matched_window_start_time_after_list)

        mims_summary_wearstate_after_list.append(matched_mims_summary_wearstate_after_list)
        wearstate_min_after_list.append(matched_wearstate_min_after_list)

    mims_summary_before_df = pd.DataFrame(list(map(np.ravel, mims_summary_before_list)))
    window_readings_before_df = pd.DataFrame(list(map(np.ravel, window_readings_before_list)))
    window_start_time_before_df = pd.DataFrame(list(map(np.ravel, window_start_time_before_list)))

    mims_summary_after_df = pd.DataFrame(list(map(np.ravel, mims_summary_after_list)))
    window_readings_after_df = pd.DataFrame(list(map(np.ravel, window_readings_after_list)))
    window_start_time_after_df = pd.DataFrame(list(map(np.ravel, window_start_time_after_list)))

    mims_summary_wearstate_after_df = pd.DataFrame(list(map(np.ravel, mims_summary_wearstate_after_list)))
    wearstate_min_after_df = pd.DataFrame(list(map(np.ravel, wearstate_min_after_list)))

    df = pd.DataFrame()
    for sec_before in sec_before_list:
        col = int(sec_before // 60 - 1)
        df = pd.concat([df, mims_summary_before_df[col].rename("MIMS_SUMMARY_" + str(sec_before // 60) + "MIN_BEFORE")],
                       axis=1)
        df = pd.concat(
            [df, window_readings_before_df[col].rename("WINDOW_NUM_READINGS_" + str(sec_before // 60) + "MIN_BEFORE")],
            axis=1)
        df = pd.concat(
            [df, window_start_time_before_df[col].rename("WINDOW_STARTTIME_" + str(sec_before // 60) + "MIN_BEFORE")],
            axis=1)

    for sec_after in sec_after_list:
        col = int(sec_after // (30 * 60) - 1)
        # overall mims
        df = pd.concat([df, mims_summary_after_df[col].rename("MIMS_SUMMARY_" + str(sec_after // 60) + "MIN_AFTER")],
                       axis=1)
        df = pd.concat(
            [df, window_readings_after_df[col].rename("WINDOW_NUM_READINGS_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)
        df = pd.concat(
            [df, window_start_time_after_df[col].rename("WINDOW_STARTTIME_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)

        # mims in wear, sleep, nonwear
        df = pd.concat(
            [df, mims_summary_wearstate_after_df[col * 3 + 0].rename(
                "MIMS_SUMMARY_WEAR_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)
        df = pd.concat(
            [df, wearstate_min_after_df[col * 3 + 0].rename("MIMS_WEAR_MIN_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)

        df = pd.concat(
            [df, mims_summary_wearstate_after_df[col * 3 + 1].rename(
                "MIMS_SUMMARY_SLEEP_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)
        df = pd.concat(
            [df, wearstate_min_after_df[col * 3 + 1].rename("MIMS_SLEEP_MIN_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)

        df = pd.concat(
            [df, mims_summary_wearstate_after_df[col * 3 + 2].rename(
                "MIMS_SUMMARY_NONWEAR_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)
        df = pd.concat(
            [df, wearstate_min_after_df[col * 3 + 2].rename("MIMS_NONWEAR_MIN_" + str(sec_after // 60) + "MIN_AFTER")],
            axis=1)

    return df


def create_column(prompt_local_datetime_series, intermediate_participant_path, date_in_study):
    print("\n> start generating the feature: Location ")

    # Read, parse and combine related intermediate file
    df_mims_combined = combine_intermediate_mims_file(intermediate_participant_path, date_in_study,
                                                      target_file_pattern_mims)
    df_mims_minute_combined = combine_intermediate_mims_swan_minute_file(intermediate_participant_path, date_in_study,
                                                                         target_file_pattern_mims_minute)
    df_swan_minute_combined = combine_intermediate_mims_swan_minute_file(intermediate_participant_path, date_in_study,
                                                                         target_file_pattern_swan_minute)

    if len(df_mims_combined) == 0 or len(df_mims_minute_combined) == 0 or len(df_swan_minute_combined) == 0:
        sec_before_list = [60 * x for x in list(range(1, 11))]  # 1-10 minute before prompt time
        sec_after_list = [30 * 60 * x for x in list(range(1, 7))]  # 30, 60, 90, 120, 150, 180 minute after prompt time

        column_names = []
        array_count = 0
        for sec_before in sec_before_list:
            column_names.append("MIMS_SUMMARY_" + str(sec_before // 60) + "MIN_BEFORE")
            column_names.append("WINDOW_NUM_READINGS_" + str(sec_before // 60) + "MIN_BEFORE")
            column_names.append("WINDOW_STARTTIME_" + str(sec_before // 60) + "MIN_BEFORE")
            array_count += 3


        for sec_after in sec_after_list:
            column_names.append("MIMS_SUMMARY_" + str(sec_after // 60) + "MIN_AFTER")
            column_names.append("WINDOW_NUM_READINGS_" + str(sec_after // 60) + "MIN_AFTER")
            column_names.append("WINDOW_STARTTIME_" + str(sec_after // 60) + "MIN_AFTER")

            column_names.append("MIMS_SUMMARY_WEAR_" + str(sec_after // 60) + "MIN_AFTER")
            column_names.append("MIMS_WEAR_MIN_" + str(sec_after // 60) + "MIN_AFTER")

            column_names.append("MIMS_SUMMARY_SLEEP_" + str(sec_after // 60) + "MIN_AFTER")
            column_names.append("MIMS_SLEEP_MIN_" + str(sec_after // 60) + "MIN_AFTER")

            column_names.append("MIMS_SUMMARY_NONWEAR_" + str(sec_after // 60) + "MIN_AFTER")
            column_names.append("MIMS_NONWEAR_MIN_" + str(sec_after // 60) + "MIN_AFTER")
            array_count += 9

        df = pd.DataFrame([[np.nan] * len(prompt_local_datetime_series)]*array_count).T
        df.columns = column_names
        return df

    df_mims_swan_minute_combined = pd.merge(df_mims_minute_combined, df_swan_minute_combined, how='left',
                                            left_on=['YEAR_MONTH_DAY', 'HOUR', 'MINUTE'],
                                            right_on=['YEAR_MONTH_DAY', 'HOUR', 'MINUTE'])

    # Match the combined parsed intermediate file with prompt feature data frame
    df = match_feature(prompt_local_datetime_series, df_mims_combined, df_mims_swan_minute_combined, date_in_study)

    return df


if __name__ == "__main__":
    pass
